@@ -30,12 +30,10 @@
 #include "yelp-io-channel.h"
 #include "yelp-info-parser.h"
 #include "yelp-utils.h"
+#include "yelp-debug.h"
 
-#ifdef YELP_DEBUG
-#define d(x) x
-#else
-#define d(x)
-#endif
+
+typedef struct _TagTableFix TagTableFix;
 
 GtkTreeIter *         find_real_top                      (GtkTreeModel *model, 
 							  GtkTreeIter *it);
@@ -44,7 +42,8 @@ GtkTreeIter *         find_real_sibling                  (GtkTreeModel *model,
 							  GtkTreeIter *comp);
 xmlNodePtr            yelp_info_parse_menu               (GtkTreeStore *tree,
 							  xmlNodePtr *node,
-							  gchar *page_content);
+							  gchar *page_content,
+							  gboolean notes);
 gboolean              get_menuoptions                    (gchar *line, 
 							  gchar **title, 
 							  gchar **ref, 
@@ -54,7 +53,12 @@ gboolean              resolve_frag_id                    (GtkTreeModel *model,
 							  GtkTreePath *path, 
 							  GtkTreeIter *iter,
 							  gpointer data);
-
+void                  fix_tag_table                      (gchar *offset, 
+							  gpointer page, 
+							  TagTableFix *a);
+void   		      info_process_text_notes            (xmlNodePtr *node, 
+							  gchar *content,
+							  GtkTreeStore *tree);
 /* Part 1: Parse File Into Tree Store */
 
 enum
@@ -92,28 +96,39 @@ page_type (char *page)
 static char
 *open_info_file (char *file)
 {
-	GIOChannel *channel;
+	GIOChannel *channel = NULL;
 	int i;
 	gsize len;
 	char *str;
 	GError *error = NULL;
 	GIOStatus result = G_IO_STATUS_NORMAL;
 
-	d (g_print ("!! Opening %s...\n", file));
+	debug_print (DB_DEBUG, "!! Opening %s...\n", file);
 	
 	channel = yelp_io_channel_new_file (file, &error);
+	/* TODO: Actually handle the errors sanely.  Don't crash */
+	if (!channel) {
+	  g_error ("Error opening file %s: %s\n", file, error->message);
+	  g_error_free (error);
+	  error = NULL;
+	  exit (655);
+	}
 	result = g_io_channel_read_to_end (channel, &str, &len, &error);
+	/* TODO: Ditto above */
 	if (result != G_IO_STATUS_NORMAL) {
 	  g_error ("Error reading file: %s\n", error->message);
+	  g_error_free (error);
+	  error = NULL;
 	  exit (666);
 	}
 	g_io_channel_shutdown (channel, FALSE, NULL);
+	g_io_channel_unref (channel);
 
 	for (i = 0; i < (len - 1); i++)
 	{
 		if (str[i] == '\0' && str[i+1] == '\b')
 		{
-		  d (g_print ("=> got a NULL, replacing\n"));
+		  debug_print (DB_WARN, "=> got a NULL, replacing\n");
 		  str[i] = ' '; str[i+1] = ' ';
 		}
 	}
@@ -122,58 +137,42 @@ static char
 }
 
 static gchar *
-find_info_part (gchar *part_name)
+find_info_part (gchar *part_name, gchar *base)
 {
-  gchar ** paths = yelp_get_info_paths ();
-  gint i;
-  gchar *filename = NULL;
-  gchar *uri = NULL;
-  GDir *dir;
+  /* New and improved.  We now assume that all parts are
+   * in the same subdirectory as the base file.  Makes
+   * life much simpler and is (afaict) always true
+   */
+  gchar *path;
+  gchar *tmp;
   gchar *bzfname, *gzfname;
+  gchar *uri = NULL;
+  tmp = g_strrstr (base, "/");
+  path = g_strndup (base, tmp-base);
 
-  bzfname = g_strconcat (part_name, ".bz2", NULL);
-  gzfname = g_strconcat (part_name, ".gz", NULL);
+  bzfname = g_strconcat (path, "/", part_name, ".bz2", NULL);
+  gzfname = g_strconcat (path, "/", part_name, ".gz", NULL);
+  
+  if (g_file_test (bzfname, G_FILE_TEST_EXISTS))
+    uri = g_strdup (bzfname);
+  else if (g_file_test (gzfname, G_FILE_TEST_EXISTS))
+    uri = g_strdup (gzfname);
 
-  for (i=0; paths[i]; i++) {
-    dir = g_dir_open (paths[i], 0, NULL);
-    if (!dir)
-      continue;
-
-    while ((filename = (gchar *) g_dir_read_name (dir))) {
-      if (g_str_equal (filename, bzfname)) {
-	uri = g_strconcat (paths[i], "/", bzfname, NULL);
-	g_dir_close (dir);
-	goto done;
-      } else if (g_str_equal (filename, gzfname)) {
-	uri = g_strconcat (paths[i], "/", gzfname, NULL);
-	g_dir_close (dir);
-	goto done;
-      } else if (g_str_equal (filename, part_name)) {
-	uri = g_strconcat (paths[i], "/", part_name, NULL);
-	g_dir_close (dir);
-	goto done;
-
-      }
-    }
-    g_dir_close (dir);
-  }
-
- done:
-  g_free (gzfname);
   g_free (bzfname);
+  g_free (gzfname);
+  g_free (path);
   return uri;
 
 }
 
 static char
-*process_indirect_map (char *page)
+*process_indirect_map (char *page, gchar * file)
 {
 	char **lines;
 	char **ptr;
-	char *composite;
+	char *composite = NULL;
 	
 	lines = g_strsplit (page, "\n", 0);
-	composite = NULL;
 
 	for (ptr = lines + 1; *ptr != NULL; ptr++);
 	for (ptr--; ptr != lines; ptr--)
@@ -185,33 +184,32 @@ static char
 		int offset;
 		int plength;
 
-		d (g_print ("Line: %s\n", *ptr));
+		debug_print (DB_DEBUG, "Line: %s\n", *ptr);
 		items = g_strsplit (*ptr, ": ", 2);
 
 		if (items[0])
 		{
-		        filename = find_info_part (items[0]);
-			str = open_info_file (filename);
-	
+		  filename = find_info_part (items[0], file);
+		  str = open_info_file (filename);
+		  
 			pages = g_strsplit (str, "", 2);
 			g_free (str);
 
 			offset =  atoi(items[1]);
 			plength = strlen(pages[1]);
 			
-			d (g_print ("Need to make string %s+%i bytes = %i\n",
+			debug_print (DB_DEBUG, "Need to make string %s+%i bytes = %i\n",
 				    items[1], plength,
-				    offset + plength));
+				    offset + plength);
 			
 			if (!composite) /* not yet created, malloc it */
 			{
 				int length;
-
 				length = offset + plength;
 				composite = g_malloc (sizeof (char) *
-						length + 1);
+						      (length + 1));
 				memset (composite, '-', length);
-				composite[length + 1] = '\0';
+				composite[length] = '\0';
 			}
 			composite[offset] = '';
 			memcpy (composite + offset + 1, pages[1], plength);
@@ -238,7 +236,8 @@ static GHashTable
 	char **ptr;
 	char **items;
 
-	table = g_hash_table_new (g_str_hash, g_str_equal);
+	table = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, 
+				       g_free);
 	lines = g_strsplit (page, "\n", 0);
 
 	for (ptr = lines; *ptr != NULL; ptr++)
@@ -246,8 +245,8 @@ static GHashTable
 		if (strncmp (*ptr, "Node: ", 6) == 0)
 		{
 			items = g_strsplit (*ptr, "", 2);
-			d (g_print ("Node: %s Offset: %s\n",
-				    items[0] + 6, items[1]));
+			debug_print (DB_DEBUG, "Node: %s Offset: %s\n",
+				    items[0] + 6, items[1]);
 			g_hash_table_insert (table,
 					g_strdup (items[0] + 6),
 					g_strdup (items[1]));
@@ -290,31 +289,12 @@ static char
 static int
 node2page (GHashTable *nodes2offsets, GHashTable *offsets2pages, char *node)
 {
-	char *offset;
-	gint int_offset;
+	char *offset = NULL;
 	gint page;
 
 	offset = g_hash_table_lookup (nodes2offsets, node);
 	page = GPOINTER_TO_INT (g_hash_table_lookup (offsets2pages, offset));
-	if (!page) {
-	  /* We got a badly formed tag table.  Most probably bash info page :(
-	   * Have to find the correct page.
-	   * The bash info file assumess 3 bytes more per node than reality
-	   * hence we check backwards in steps of 3
-	   * The first one we come across should be the correct node
-	   * (fingers crossed at least)
-	   */
-	  gchar *new_offset = NULL;
-	  int_offset = atoi (offset);
 
-	  while (!page && int_offset > 0) {
-	    int_offset-=3;
-	    g_free (new_offset);
-	    new_offset = g_strdup_printf ("%d", int_offset);
-	    page = GPOINTER_TO_INT (g_hash_table_lookup (offsets2pages, new_offset));	    
-	  }
-	  g_free (new_offset);
-	}
 	return page;
 }
 
@@ -324,7 +304,7 @@ static GtkTreeIter
 	GtkTreeIter *iter;
 
 	iter = g_hash_table_lookup (nodes2iters, node);
-	d (if (!iter) g_print ("Could not retrieve iter for node !%s!\n", node));
+	d (if (!iter) debug_print (DB_WARN, "Could not retrieve iter for node !%s!\n", node));
 	return iter;
 }
 
@@ -340,6 +320,7 @@ GtkTreeIter
   r = gtk_tree_iter_copy (it);
   tmp = g_malloc0 (sizeof (GtkTreeIter));
   while (gtk_tree_model_iter_parent (model, tmp, r)) {
+    gtk_tree_iter_free (r);
     r = gtk_tree_iter_copy (tmp);
   }
   g_free (tmp);
@@ -356,8 +337,9 @@ GtkTreeIter * find_real_sibling (GtkTreeModel *model,
   gchar *title;
   gchar *reftitle;
 
-  if (!it)
+  if (!it) {
     return NULL;
+  }
 
   r = gtk_tree_iter_copy (it);
   tmp = gtk_tree_iter_copy (it);
@@ -378,7 +360,7 @@ GtkTreeIter * find_real_sibling (GtkTreeModel *model,
       title = gtk_tree_model_get_string_from_iter (model, r);
   }
 
-  g_free (r);
+  gtk_tree_iter_free (r);
   g_free (title);
   g_free (reftitle);
   return tmp;
@@ -397,6 +379,7 @@ process_page (GtkTreeStore *tree, GHashTable *nodes2offsets,
 	char *up;
 	char *prev;
 	char *next;
+	gchar *tmp;
 
 	int page;
 	
@@ -419,11 +402,12 @@ process_page (GtkTreeStore *tree, GHashTable *nodes2offsets,
 
 	/* check to see if this page has been processed already */
 	page = node2page (nodes2offsets, offsets2pages, node);
-	if (processed_table[page])
+	if (processed_table[page]) {
 		return;
+	}
 	processed_table[page] = 1;
 	
-	d (g_print ("-- Processing Page %s\n\tParent: %s\n", node, up));
+	debug_print (DB_DEBUG, "-- Processing Page %s\n\tParent: %s\n", node, up);
 
 	iter = g_malloc0 (sizeof (GtkTreeIter));
 	/* check to see if we need to process our parent and siblings */
@@ -432,7 +416,7 @@ process_page (GtkTreeStore *tree, GHashTable *nodes2offsets,
 		page = node2page (nodes2offsets, offsets2pages, up);
 		if (!processed_table[page])
 		{
-		  d (g_print ("%% Processing Node %s\n", up));
+		  debug_print (DB_DEBUG, "%% Processing Node %s\n", up);
 			process_page (tree, nodes2offsets, offsets2pages,
 				nodes2iters, processed_table, page_list,
 				page_list[page]);
@@ -446,7 +430,7 @@ process_page (GtkTreeStore *tree, GHashTable *nodes2offsets,
 	      page = node2page (nodes2offsets, offsets2pages, prev);
 	      if (!processed_table[page])
 		{
-		  d (g_print ("%% Processing Node %s\n", prev));
+		  debug_print (DB_DEBUG, "%% Processing Node %s\n", prev);
 		  process_page (tree, nodes2offsets, offsets2pages,
 				nodes2iters, processed_table, page_list,
 				page_list[page]);
@@ -457,10 +441,10 @@ process_page (GtkTreeStore *tree, GHashTable *nodes2offsets,
 	/* by this point our parent and older sibling should be processed */
 	if (!up || !g_ascii_strcasecmp (up, "(dir)") || !strcmp (up, "Top"))
 	{
-	  d (g_print ("\t> no parent\n"));
+	  debug_print (DB_DEBUG, "\t> no parent\n");
 		if (!prev || !g_ascii_strcasecmp (prev, "(dir)"))
 		{
-		  d (g_print ("\t> no previous\n"));
+		  debug_print (DB_DEBUG, "\t> no previous\n");
 			gtk_tree_store_append (tree, iter, NULL);
 		}
 		else if (prev) {
@@ -470,7 +454,7 @@ process_page (GtkTreeStore *tree, GHashTable *nodes2offsets,
 		  if (real) {
 		    gtk_tree_store_insert_after (tree, iter, NULL,
 						 real);
-		    g_free (real);
+		    gtk_tree_iter_free (real);
 		  }
 		  else 
 		    gtk_tree_store_append (tree, iter, NULL);
@@ -478,7 +462,7 @@ process_page (GtkTreeStore *tree, GHashTable *nodes2offsets,
 	}
 	else if (!prev || !g_ascii_strcasecmp (prev, "(dir)") || !strcmp (prev, up))
 	{
-	  d (g_print ("\t> no previous\n"));
+	  debug_print (DB_DEBUG, "\t> no previous\n");
 		gtk_tree_store_append (tree, iter,
 			node2iter (nodes2iters, up));
 	}
@@ -486,37 +470,40 @@ process_page (GtkTreeStore *tree, GHashTable *nodes2offsets,
 	{
 	  GtkTreeIter *upit = node2iter (nodes2iters, up);
 	  GtkTreeIter *previt = node2iter (nodes2iters, prev);
-	  GtkTreeIter *nit;
-	  d (g_print ("+++ Parent: %s Previous: %s\n", up, prev));
+	  GtkTreeIter *nit = NULL;
+	  debug_print (DB_DEBUG, "+++ Parent: %s Previous: %s\n", up, prev);
 	  
-	  d (if (upit) g_print ("++++ Have parent node!\n"));
-	  d (if (previt) g_print ("++++ Have previous node!\n"));
+	  d (if (upit) debug_print (DB_DEBUG, "++++ Have parent node!\n"));
+	  d (if (previt) debug_print (DB_DEBUG, "++++ Have previous node!\n"));
 	  nit = find_real_sibling (GTK_TREE_MODEL (tree), previt, upit);
 	  if (nit) {
 	    gtk_tree_store_insert_after (tree, iter,
 					 upit,
 					 nit);
-	    g_free (nit);
+	    gtk_tree_iter_free (nit);
 	  }
 	  else
 	    gtk_tree_store_append (tree, iter, upit);
 	}
 	else
 	{
-	  d (g_print ("# node %s was not put in tree\n", node));
+	  debug_print (DB_DEBUG, "# node %s was not put in tree\n", node);
 	  return;
 	}
 
-	d (if (iter) g_print ("Have a valid iter, storing for %s\n", node));
+	d (if (iter) debug_print (DB_DEBUG, "Have a valid iter, storing for %s\n", node));
 	g_hash_table_insert (nodes2iters, g_strdup (node), iter);
-	d (g_print ("size: %i\n", g_hash_table_size (nodes2iters)));
-	gtk_tree_store_set (tree, iter,
-			COLUMN_PAGE_NO, g_strdup_printf ("%i",
-				node2page (nodes2offsets, offsets2pages, node)),
-			COLUMN_PAGE_NAME, g_strdup (node),
-			COLUMN_PAGE_CONTENT, g_strdup (parts[2]),
-			-1);
+	debug_print (DB_DEBUG, "size: %i\n", g_hash_table_size (nodes2iters));
 
+	tmp = g_strdup_printf ("%i",
+			       node2page (nodes2offsets, offsets2pages, node));
+	gtk_tree_store_set (tree, iter,
+			    COLUMN_PAGE_NO, tmp,
+			    COLUMN_PAGE_NAME, node,
+			    COLUMN_PAGE_CONTENT, parts[2],
+			    -1);
+
+	g_free (tmp);
 	g_free (node);
 	g_free (up);
 	g_free (prev);
@@ -524,6 +511,31 @@ process_page (GtkTreeStore *tree, GHashTable *nodes2offsets,
 	g_strfreev (parts);
 }
 
+/* These are used to fix the tag tables to the correct offsets.
+ * Assuming out offsets are correct (because we calculated them, these values
+ * are used to overwrite the offsets declared in the info file */
+struct _TagTableFix {
+  GHashTable *nodes2offsets;
+  GHashTable *pages2nodes;
+};
+
+void
+fix_tag_table (gchar *offset, gpointer page, TagTableFix *a)
+{
+  gchar *node_name = NULL;
+
+  node_name = g_hash_table_lookup (a->pages2nodes, page);
+
+  if (!node_name) 
+    return;
+
+  g_hash_table_replace (a->nodes2offsets, g_strdup (node_name), g_strdup (offset));
+
+
+
+}
+
+
 /**
  * Parse file into a GtkTreeStore containing useful information that we can
  * later convert into a nice XML document or something else.
@@ -536,33 +548,43 @@ GtkTreeStore
 	int pages;
 	int offset;
 	GHashTable *nodes2offsets = NULL;
-	GHashTable *offsets2pages;
-	GHashTable *nodes2iters;
+	GHashTable *offsets2pages = NULL;
+	GHashTable *pages2nodes = NULL;
+	GHashTable *nodes2iters = NULL;
 	int *processed_table;
 	GtkTreeStore *tree;
 	int pt;
-	char *str;
+	char *str = NULL;
 	gboolean chained_info;
+	TagTableFix *ttf;
 	
 	str = open_info_file (file);
 	page_list = g_strsplit (str, "\n", 0);
 
 	g_free (str);
+	str = NULL;
 	
 	pages = 0;
 	offset = 0;
 	chained_info = FALSE;
 
-	offsets2pages = g_hash_table_new (g_str_hash, g_str_equal);
-	
+	offsets2pages = g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
+					       NULL);
+	pages2nodes = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, 
+					     g_free);
+
 	for (ptr = page_list; *ptr != NULL; ptr++)
 	{
-	  d (g_print ("page %i at offset %i\n", pages, offset));
-/*		g_print ("page starts:\n%s...\n", *ptr); */
+	  gchar *name = NULL;
+	  debug_print (DB_DEBUG, "page %i at offset %i\n", pages, offset);
 
 		g_hash_table_insert (offsets2pages,
 				g_strdup_printf ("%i", offset), 
 				GINT_TO_POINTER (pages));
+		name = get_value_after (*ptr, "Node: ");
+		if (name)
+		  g_hash_table_insert (pages2nodes,
+				       GINT_TO_POINTER (pages), name);
 		
 		offset += strlen (*ptr);
 		if (pages) offset += 2;
@@ -570,16 +592,16 @@ GtkTreeStore
 		pt = page_type (*ptr);
 		if (pt == PAGE_TAG_TABLE)
 		{
-		  d (g_print ("Have the Tag Table\n"));
+		  debug_print (DB_DEBUG, "Have the Tag Table\n");
 			/* this needs to be freed later too */
 			nodes2offsets = process_tag_table (*ptr);
 			break;
 		}
 		else if (pt == PAGE_INDIRECT)
 		{
-		  d (g_print ("Have the indirect mapping table\n"));
+		  debug_print (DB_DEBUG, "Have the indirect mapping table\n");
 			chained_info = TRUE;
-			str = process_indirect_map (*ptr);
+			str = process_indirect_map (*ptr, file);
 		}
 	}
 
@@ -589,18 +611,20 @@ GtkTreeStore
 		 * more processing */
 		g_strfreev (page_list);
 		g_hash_table_destroy (offsets2pages);
+		offsets2pages = g_hash_table_new_full (g_str_hash, 
+						       g_str_equal, g_free,
+						       NULL);
 		
 		pages = 0;
 		offset = 0;
-		
+
 		page_list = g_strsplit (str, "\n", 0);
-		offsets2pages = g_hash_table_new (g_str_hash, g_str_equal);
 		
 		g_free (str);
 		
 		for (ptr = page_list; *ptr != NULL; ptr++)
 		{
-		  d (g_print ("page %i at offset %i\n", pages, offset));
+		  debug_print (DB_DEBUG, "page %i at offset %i\n", pages, offset);
 			g_hash_table_insert (offsets2pages,
 					g_strdup_printf ("%i", offset),
 					 GINT_TO_POINTER (pages));
@@ -611,6 +635,16 @@ GtkTreeStore
 	}
 	if (!nodes2offsets)
 	  return NULL;
+	/* We now go through the offsets2pages dictionary and correct the entries
+	 * as the tag tables are known to lie.  Yes, they do.  Consistantly and
+	 * maliciously
+	 */	
+	ttf = g_new0 (TagTableFix, 1);
+	ttf->nodes2offsets = nodes2offsets;
+	ttf->pages2nodes = pages2nodes;
+	g_hash_table_foreach (offsets2pages, (GHFunc) fix_tag_table, ttf);
+	g_free (ttf);
+
 
 	/* at this point we have two dictionaries,
 	 * 	node names:offsets, and
@@ -620,7 +654,8 @@ GtkTreeStore
 	processed_table = g_malloc0 (pages * sizeof (int));
 	tree = gtk_tree_store_new (N_COLUMNS, G_TYPE_STRING, G_TYPE_STRING,
 			G_TYPE_STRING);
-	nodes2iters = g_hash_table_new (g_str_hash, g_str_equal);
+	nodes2iters = g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
+					     (GDestroyNotify) gtk_tree_iter_free);
 
 	pages = 0;
 	for (ptr = page_list; *ptr != NULL; ptr++)
@@ -648,11 +683,12 @@ parse_tree_level (GtkTreeStore *tree, xmlNodePtr *node, GtkTreeIter iter)
 	GtkTreeIter children;
 	xmlNodePtr newnode;
 
-	char *page_no;
-	char *page_name;
-	char *page_content;
-	
-	d (g_print ("Decended\n"));
+	char *page_no = NULL;
+	char *page_name = NULL;
+	char *page_content = NULL;
+	gboolean notes = FALSE;
+
+	debug_print (DB_DEBUG, "Decended\n");
 	do
 	{
 		gtk_tree_model_get (GTK_TREE_MODEL (tree), &iter,
@@ -660,17 +696,26 @@ parse_tree_level (GtkTreeStore *tree, xmlNodePtr *node, GtkTreeIter iter)
 				COLUMN_PAGE_NAME, &page_name,
 				COLUMN_PAGE_CONTENT, &page_content,
 				-1);
-		d (g_print ("Got Section: %s\n", page_name));
-		if (strstr (page_content, "\n* Menu:")) {
-		  newnode = yelp_info_parse_menu (tree, node, page_content);
+		debug_print (DB_DEBUG, "Got Section: %s\n", page_name);
+		if (strstr (page_content, "*Note") || 
+		    strstr (page_content, "*note")) {
+		  notes = TRUE;
+		}
+		if (strstr (page_content, "* Menu:")) {
+		  newnode = yelp_info_parse_menu (tree, node, page_content, notes);
 		} else {
 		  newnode = xmlNewTextChild (*node, NULL,
 					     BAD_CAST "Section",
 					     NULL);
-		  xmlNewTextChild (newnode, NULL,
-				   BAD_CAST "para",
-				   BAD_CAST page_content);
+		  if (!notes)
+		    xmlNewTextChild (newnode, NULL,
+				     BAD_CAST "para",
+				     BAD_CAST page_content);
 		  
+		  else {
+		    /* Handle notes here */
+		    info_process_text_notes (&newnode, page_content, tree);
+		  }
 		}
 		/* if we free the page content, now it's in the XML, we can
 		 * save some memory */
@@ -678,16 +723,18 @@ parse_tree_level (GtkTreeStore *tree, xmlNodePtr *node, GtkTreeIter iter)
 		page_content = NULL;
 
 		xmlNewProp (newnode, BAD_CAST "id", 
-			    BAD_CAST g_strdup (page_no));
+			    BAD_CAST page_no);
 		xmlNewProp (newnode, BAD_CAST "name", 
-			    BAD_CAST g_strdup (page_name));
+			    BAD_CAST page_name);
 		if (gtk_tree_model_iter_children (GTK_TREE_MODEL (tree),
 				&children,
 				&iter))
-			parse_tree_level (tree, &newnode, children);
+		  parse_tree_level (tree, &newnode, children);
+		g_free (page_no);
+		g_free (page_name);
 	}
 	while (gtk_tree_model_iter_next (GTK_TREE_MODEL (tree), &iter));
-	d (g_print ("Ascending\n"));
+	debug_print (DB_DEBUG, "Ascending\n");
 }
 
 xmlDocPtr
@@ -714,7 +761,7 @@ yelp_info_parser_parse_tree (GtkTreeStore *tree)
 
 	if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (tree), &iter))
 		parse_tree_level (tree, &node, iter);
-	d (else g_print ("Empty tree?\n"));
+	d (else debug_print (DB_DEBUG, "Empty tree?\n"));
 
 	/*
 	xmlDocDumpFormatMemory (doc, &xmlbuf, &bufsiz, 1);
@@ -724,24 +771,27 @@ yelp_info_parser_parse_tree (GtkTreeStore *tree)
 	return doc;
 }
 
-static gchar *section_id;
-
 gboolean
 resolve_frag_id (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter,
 		 gpointer data)
 {
-  gchar *page_no;
-  gchar *page_name;
+  gchar *page_no = NULL;
+  gchar *page_name = NULL;
+  gchar **xref = data;
 
   gtk_tree_model_get (GTK_TREE_MODEL (model), iter,
 		      COLUMN_PAGE_NO, &page_no,
 		      COLUMN_PAGE_NAME, &page_name,
 		      -1);
-
-  if (g_str_equal (page_name, data)) {
-    section_id = g_strdup (page_no);
+  if (g_str_equal (page_name, *xref)) {
+    g_free (*xref);
+    *xref = g_strdup (page_no);
+    g_free (page_name);
+    g_free (page_no);
     return TRUE;
   }
+  g_free (page_name);
+  g_free (page_no);
 
   return FALSE;
 }
@@ -779,7 +829,6 @@ get_menuoptions (gchar *line, gchar **title, gchar **ref, gchar **desc,
 
     tfind+=2;
     (*desc) = g_strdup (tfind);
-
   } else { /* The other type of menu option */
     gchar *td = NULL;
 
@@ -799,24 +848,31 @@ get_menuoptions (gchar *line, gchar **title, gchar **ref, gchar **desc,
 
 xmlNodePtr
 yelp_info_parse_menu (GtkTreeStore *tree, xmlNodePtr *node, 
-		      gchar *page_content)
+		      gchar *page_content, gboolean notes)
 {
   gchar **split;
   gchar **menuitems;
+  gchar *tmp = NULL;
   xmlNodePtr newnode;
   int i=0;
 
-  split = g_strsplit (page_content, "\n* Menu:", 2);
-
+  split = g_strsplit (page_content, "* Menu:", 2);
+  
   newnode = xmlNewChild (*node, NULL,
 			 BAD_CAST "Section", NULL);
+    
 
-  xmlNewTextChild (newnode, NULL,
-	       BAD_CAST "para", BAD_CAST g_strconcat (split[0],
-						      "\n* Menu:",
-						      NULL));
+  tmp = g_strconcat (split[0], "\n* Menu:", NULL);
+  if (!notes)
+    xmlNewTextChild (newnode, NULL,
+		     BAD_CAST "para", BAD_CAST tmp);
+  else {
+    info_process_text_notes (&newnode, tmp, tree);
+  }
+  g_free (tmp);
 
   menuitems = g_strsplit (split[1], "\n", -1);
+  g_strfreev (split);
 
   while (menuitems[i] != NULL) {
     gboolean menu = FALSE;
@@ -831,14 +887,16 @@ yelp_info_parse_menu (GtkTreeStore *tree, xmlNodePtr *node,
  
     if (menu) {
       mholder = xmlNewChild (newnode, NULL, BAD_CAST "menuholder", NULL);
-      gtk_tree_model_foreach (GTK_TREE_MODEL (tree), resolve_frag_id, xref);
+      gtk_tree_model_foreach (GTK_TREE_MODEL (tree), resolve_frag_id, &xref);
       
       if (ref == NULL) { /* A standard type menu */
+	tmp = g_strconcat (title, "::", NULL);
 	ref1 = xmlNewTextChild (mholder, NULL, BAD_CAST "a",
-					BAD_CAST g_strconcat (title, "::", 
-							      NULL));
-	xmlNewProp (ref1, BAD_CAST "href", BAD_CAST g_strconcat ("#", section_id, 
-							      NULL));
+				BAD_CAST tmp);
+	g_free (tmp);
+	tmp = g_strconcat ("#", xref, NULL);
+	xmlNewProp (ref1, BAD_CAST "href", BAD_CAST tmp);
+	g_free (tmp);
       } else { /* Indexy type menu  - we gotta do a  little work to fix the
 		* spacing
 		*/
@@ -854,16 +912,19 @@ yelp_info_parse_menu (GtkTreeStore *tree, xmlNodePtr *node,
 	
 	ref1 = xmlNewTextChild (mholder, NULL, BAD_CAST "a",
 					BAD_CAST title);
-	xmlNewProp (ref1, BAD_CAST "href", BAD_CAST g_strconcat ("#", section_id, 
-							      NULL));
+	tmp = g_strconcat ("#", xref, NULL);
+	xmlNewProp (ref1, BAD_CAST "href", BAD_CAST tmp);
+	g_free (tmp);
 	xmlNewTextChild (mholder, NULL, BAD_CAST "spacing",
 			 BAD_CAST sp);
+	tmp = g_strconcat (g_strstrip(ref), ".", NULL);
 	ref1 = xmlNewTextChild (mholder, NULL, BAD_CAST "a",
-					 BAD_CAST g_strconcat (g_strstrip(ref), 
-							       ".", NULL));
-	xmlNewProp (ref1, BAD_CAST "href", BAD_CAST g_strconcat ("#", section_id, 
-							       NULL));
-	
+				BAD_CAST tmp);
+	g_free (tmp);
+	tmp = g_strconcat ("#", xref, NULL);
+	xmlNewProp (ref1, BAD_CAST "href", BAD_CAST tmp);
+
+	g_free (tmp);	
 	g_free (sp);
       }
       xmlNewTextChild (mholder, NULL, BAD_CAST "para",
@@ -874,8 +935,194 @@ yelp_info_parse_menu (GtkTreeStore *tree, xmlNodePtr *node,
       
     }
     i++;
-    
+    g_free (title);
+    g_free (ref);
+    g_free (desc);
+    g_free (xref);
+
   }
+  g_strfreev (menuitems);
   
   return newnode;
 }
+
+void
+info_process_text_notes (xmlNodePtr *node, gchar *content, GtkTreeStore *tree)
+{
+  gchar **notes;
+  gchar **current;
+  xmlNodePtr holder;
+  xmlNodePtr ref1;
+  gboolean first = TRUE;
+
+  notes = g_strsplit (content, "*Note", -1);
+  holder = xmlNewChild (*node, NULL, BAD_CAST "noteholder", NULL);
+
+  for (current = notes; *current != NULL; current++) {
+    /* Since the notes can be either *Note or *note, we handle the second 
+     * variety here
+     */
+    gchar **subnotes;
+    gchar **current_real;
+
+    subnotes = g_strsplit (*current, "*note", -1);
+    for (current_real = subnotes; *current_real != NULL; current_real++) {
+      gchar *url, **urls, **ulink;
+      gchar *append;
+      gchar *alt_append, *alt_append1;
+      gchar *link_text;
+      gchar *href = NULL;
+      gchar *break_point = NULL;
+      if (first) {
+	/* The first node is special.  It doesn't have a note ref at the 
+	 * start, so we can just add it and forget about it.
+	 */
+	first = FALSE;
+	xmlNewTextChild (holder, NULL, BAD_CAST "para1",
+			 BAD_CAST (*current_real));
+	continue;
+      }
+      /* If we got to here, we now gotta parse the note reference */
+
+      append = strchr (*current_real, ':');
+      if (!append) {
+      xmlNewTextChild (holder, NULL, BAD_CAST "para1",
+		       BAD_CAST *current_real);
+      continue;
+      }
+      append++;
+      alt_append = append;
+      alt_append1 = alt_append;
+      append = strchr (append, ':');
+      alt_append = strchr (alt_append, '.');
+      alt_append1 = strchr (alt_append1, ',');
+      if (!append && !alt_append && !alt_append1) {
+	xmlNewTextChild (holder, NULL, BAD_CAST "para1",
+			 BAD_CAST *current_real);
+	continue;
+      }
+      if (!append || alt_append || alt_append1) {
+	if (!append) {
+	  if (alt_append) append = alt_append;
+	  else append = alt_append1;
+	}
+	if (alt_append && alt_append < append)
+	  append = alt_append;
+	if (alt_append1 && alt_append1 < append)
+	  append = alt_append1;
+      }
+      append++;
+
+      url = g_strndup (*current_real, append - (*current_real));
+      /* By now, we got 2 things.  First, is append which is the (hopefully)
+       * non-link text.  Second, we got a url.
+       * The url can be in several forms:
+       * 1. linkend::
+       * 2. linkend:(infofile)Linkend.
+       * 3. linkend: infofile Linkend.
+       * 4. linkend: (infofile)Linkend, (pretty sure this is just broken)
+       * All possibilities should have been picked up.
+       * Here:
+       * Clean up the split.  Should be left with a real url and
+       * a list of fragments that should be linked
+       * Also goes through and removes extra spaces, leaving only one 
+       * space in place of many
+       */
+      urls = g_strsplit (url, "\n", -1);
+      break_point = strchr (url, '\n');
+      while (break_point) {
+	*break_point = ' ';
+	break_point = strchr (++break_point, '\n');
+      }
+      break_point = strchr (url, ' ');
+      while (break_point) {
+	if (*(break_point+1) == ' ') {
+	  /* Massive space.  Fix. */
+	  gchar *next = break_point;
+	  gchar *url_copy;
+	  while (*next == ' ')
+	    next++;
+	  next--;
+	  url_copy = g_strndup (url, break_point-url);
+	  g_free (url);
+	  url = g_strconcat (url_copy, next, NULL);
+	  break_point = strchr (url, ' ');
+	  g_free (url_copy);
+	} else {
+	  break_point++;
+	  break_point = strchr (break_point, ' ');
+	}
+      }
+      if (url[strlen(url)-1] == '.') { /* The 2nd or 3rd sort of link */ 
+	gchar *stop = NULL;
+	gchar *lurl = NULL;
+	gchar *zloc = NULL;
+	stop = strchr (url, ':');
+	lurl = strchr (stop, '(');
+	if (!lurl) { /* 3rd type of link */
+	  gchar **t;
+	  stop++;
+	  t = g_strsplit (stop, " ", 3);
+	  zloc = &(t[2][strlen(t[2])-1]);
+	  *zloc = '\0';
+	  href = g_strconcat ("info:", t[1], "#", t[2], NULL);
+	  *zloc = 'a';
+	  g_strfreev (t);
+	} else { /* 2nd type of link.  Easy. */
+	  zloc = &(lurl[strlen(lurl)-1]);
+	  *zloc = '\0';
+	  href = g_strconcat ("info:", lurl, NULL);
+	  *zloc = 'a';
+	}
+      } else { /* First kind of link */
+	gchar *tmp1;
+	gchar *frag;
+
+	tmp1 = strchr (url, ':');
+	frag = g_strndup (url, tmp1 - url);
+	g_strstrip (frag);
+	gtk_tree_model_foreach (GTK_TREE_MODEL (tree), resolve_frag_id, &frag);
+	href = g_strconcat ("#", frag, NULL);
+	g_free (frag);
+      }
+      for (ulink = urls; *ulink != NULL; ulink++) {
+	if (ulink == urls)
+	  link_text = g_strconcat ("*Note", *ulink, NULL);
+	else {
+	  gchar *spacing = *ulink;
+	  gchar *tmp;
+	  while (*spacing == ' ') {
+	    spacing++;
+	  }
+	  if (spacing != *ulink) {
+	    spacing-=2;
+	    tmp = g_strndup (*ulink, spacing-*ulink);
+	    spacing+=2;
+	    xmlNewTextChild (holder, NULL, BAD_CAST "spacing",
+			     BAD_CAST tmp);
+	    g_free (tmp);
+	    link_text = g_strdup (spacing);
+	  } else {
+	    link_text = g_strdup (*ulink);
+	  }
+	}
+	ref1 = xmlNewTextChild (holder, NULL, BAD_CAST "a",
+				BAD_CAST link_text);
+	if (*(ulink+1) != NULL)
+	  xmlNewTextChild (holder, NULL, BAD_CAST "para",
+			   BAD_CAST "");
+
+	g_free (link_text);
+	xmlNewProp (ref1, BAD_CAST "href", BAD_CAST href);
+      }
+      g_strfreev (urls);
+      /* Finally, we can add the text as required */
+      xmlNewTextChild (holder, NULL, BAD_CAST "para1",
+		       BAD_CAST append);
+      g_free (url);
+      g_free (href);
+    }
+    g_strfreev (subnotes);
+  }
+  g_strfreev (notes);
+}
