@@ -46,18 +46,13 @@
 #include <libxslt/xsltInternals.h>
 #include <libxslt/xsltutils.h>
 
+#include "yelp-debug.h"
 #include "yelp-error.h"
 #include "yelp-settings.h"
 #include "yelp-toc-pager.h"
 #include "yelp-utils.h"
 #include "yelp-io-channel.h"
 
-#ifdef YELP_DEBUG
-#define d(x) x
-#else
-#define d(x)
-#endif
-
 #define YELP_NAMESPACE "http://www.gnome.org/yelp/ns"
 
 #define STYLESHEET_PATH DATADIR"/yelp/xslt/"
@@ -65,8 +60,6 @@
 
 typedef gboolean      (*ProcessFunction)        (YelpTocPager      *pager);
 
-typedef struct _YelpListing YelpListing;
-
 #define YELP_TOC_PAGER_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), YELP_TYPE_TOC_PAGER, YelpTocPagerPriv))
 
 struct _YelpTocPagerPriv {
@@ -89,6 +82,7 @@ struct _YelpTocPagerPriv {
     GSList *mandir_langpath;     /* ptr to current entry in mandir_ptr */
     GHashTable *man_secthash;
     GHashTable *man_manhash;
+    GHashTable *man_dirlang;     /* key = man page directory, value = language */
 #endif
 
     xmlDocPtr     toc_doc;
@@ -107,16 +101,6 @@ struct _YelpTocPagerPriv {
     xsltTransformContextPtr transformContext;
 };
 
-struct _YelpListing {
-    gchar       *id;
-    gchar       *title;
-
-    GSList      *listings;
-    GSList      *documents;
-
-    gboolean     has_listings;
-};
-
 static void          toc_pager_class_init      (YelpTocPagerClass *klass);
 static void          toc_pager_init            (YelpTocPager      *pager);
 static void          toc_pager_dispose         (GObject           *gobject);
@@ -257,10 +241,12 @@ yelp_toc_pager_pause (YelpTocPager *pager)
 {
     g_return_if_fail (pager != NULL);
 
-    d (g_print ("yelp_toc_pager_pause\n"));
-    d (g_print ("  pause_caunt = %i\n", pager->priv->pause_count + 1));
+    debug_print (DB_FUNCTION, "entering\n");
+    debug_print (DB_ARG, "  pause_count = %i\n", pager->priv->pause_count + 1);
 
     pager->priv->pause_count = pager->priv->pause_count + 1;
+
+    debug_print (DB_FUNCTION, "exiting\n");
 }
 
 void
@@ -268,8 +254,8 @@ yelp_toc_pager_unpause (YelpTocPager *pager)
 {
     g_return_if_fail (pager != NULL);
 
-    d (g_print ("yelp_toc_pager_unpause\n"));
-    d (g_print ("  pause_caunt = %i\n", pager->priv->pause_count - 1));
+    debug_print (DB_FUNCTION, "entering\n");
+    debug_print (DB_ARG, "  pause_count = %i\n", pager->priv->pause_count - 1);
 
     pager->priv->pause_count = pager->priv->pause_count - 1;
     if (pager->priv->pause_count < 0) {
@@ -282,6 +268,8 @@ yelp_toc_pager_unpause (YelpTocPager *pager)
 			       (GtkFunction) toc_process_pending,
 			       pager);
     }
+    
+    debug_print (DB_FUNCTION, "exiting\n");
 }
 
 /******************************************************************************/
@@ -289,7 +277,7 @@ yelp_toc_pager_unpause (YelpTocPager *pager)
 static void
 toc_pager_error (YelpPager *pager)
 {
-    d (g_print ("toc_pager_error\n"));
+    debug_print (DB_FUNCTION, "entering\n");
     yelp_pager_set_state (pager, YELP_PAGER_STATE_ERROR);
 }
 
@@ -298,7 +286,7 @@ toc_pager_cancel (YelpPager *pager)
 {
     YelpTocPagerPriv *priv = YELP_TOC_PAGER (pager)->priv;
 
-    d (g_print ("toc_pager_cancel\n"));
+    debug_print (DB_FUNCTION, "entering\n");
     yelp_pager_set_state (pager, YELP_PAGER_STATE_INVALID);
 
     priv->cancel = TRUE;
@@ -307,7 +295,7 @@ toc_pager_cancel (YelpPager *pager)
 static void
 toc_pager_finish (YelpPager   *pager)
 {
-    d (g_print ("toc_pager_finish\n"));
+    debug_print (DB_FUNCTION, "entering\n");
     yelp_pager_set_state (pager, YELP_PAGER_STATE_FINISHED);
 }
 
@@ -316,7 +304,7 @@ toc_pager_process (YelpPager *pager)
 {
     YelpTocPagerPriv *priv = YELP_TOC_PAGER (pager)->priv;
 
-    d (g_print ("toc_pager_process\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     yelp_pager_set_state (pager, YELP_PAGER_STATE_PARSING);
     g_signal_emit_by_name (pager, "parse");
@@ -832,6 +820,7 @@ process_omf_pending (YelpTocPager *pager)
 
     while (filelist && filelist->data) {
 	gchar *id = NULL;
+	gchar *url = NULL;
 
 	firstfile = filelist;
 	filelist = g_slist_remove_link (filelist, firstfile);
@@ -846,6 +835,12 @@ process_omf_pending (YelpTocPager *pager)
 	if (!omf_hash)
 	    goto done;
 
+	/* url is required, if it's not present don't even add the
+	 * entry to the cache file */
+	url = g_hash_table_lookup (omf_hash, "url");
+	if (url == NULL || *url == '\0')
+	    goto done;
+
 	/* this add all the omf information to the xml cache file */
 	priv->omf_ins = xmlNewChild (priv->omf_ins, NULL, 
 			             BAD_CAST "omffile", NULL);
@@ -983,6 +978,14 @@ add_man_page_to_toc (YelpTocPager *pager, gchar *dirname, gchar *filename)
 	    info = yelp_doc_info_get (url_full, TRUE);
 
 	    if (info) {
+		gchar *lang = NULL;
+
+		if (priv->man_dirlang)
+		    lang = g_hash_table_lookup (priv->man_dirlang, dirname);
+
+		if (lang)
+		    yelp_doc_info_set_language (info, lang);
+
 		yelp_doc_info_add_uri (info, url_short, YELP_URI_TYPE_MAN);
 		tmp = xmlNewChild (tmp, NULL, BAD_CAST "doc", NULL);
 		xmlNewProp (tmp, BAD_CAST "href", BAD_CAST url_full);
@@ -1043,11 +1046,14 @@ create_manindex_file (gchar *index_file, xmlDocPtr xmldoc)
 static int
 create_toc_from_index (YelpTocPager *pager, gchar *index_file)
 {
+    const gchar * const * langs = g_get_language_names ();
     xmlXPathContextPtr xpath = NULL;
     xmlXPathObjectPtr objsect = NULL;
     xmlDocPtr manindex_xml = NULL;
+    xmlNodePtr root = NULL;
+    xmlChar *language = NULL;
     gint update_flag = 0;
-    gint i, j, k;
+    gint i, j;
     
     YelpTocPagerPriv *priv  = YELP_TOC_PAGER (pager)->priv;
 
@@ -1059,6 +1065,23 @@ create_toc_from_index (YelpTocPager *pager, gchar *index_file)
 	return 0;
     }
     
+    /* check the language that this index file was generated for */
+    root = xmlDocGetRootElement (manindex_xml);
+    language = xmlGetProp (root, BAD_CAST "lang");
+    
+    /* if the language is not the same as the current language (specified
+     * by the LANGUAGE environment variable) then return so that the index 
+     * is recreated */
+    if (language == NULL || !g_str_equal (BAD_CAST language, langs[0])) {
+	debug_print (DB_INFO, 
+	             "LANGUAGE and index file language do not match, "
+		     "index file will be recreated.\n");
+	xmlFreeDoc (manindex_xml);
+	return 0;
+    }
+
+    xmlFree (language);
+    
     xpath = xmlXPathNewContext (manindex_xml);
     objsect = xmlXPathEvalExpression (BAD_CAST "/manindex/mansect", xpath);
 
@@ -1073,28 +1096,29 @@ create_toc_from_index (YelpTocPager *pager, gchar *index_file)
 	objdirs = xmlXPathEvalExpression (BAD_CAST "dir", xpath);
 
 	for (j=0; j < objdirs->nodesetval->nodeNr; j++) {
-	    xmlXPathObjectPtr objdirname;
-	    xmlXPathObjectPtr objdirmtime;
-	    xmlXPathObjectPtr objmanpages;
 	    xmlNodePtr dirnode = objdirs->nodesetval->nodeTab[j];
 	    xmlNodePtr node = NULL;
 	    xmlChar *dirname = NULL;
 	    xmlChar *dirmtime = NULL;
+	    xmlChar *lang = NULL;
 	    time_t mtime;
 	    struct stat buf;
 	    
-	    xpath->node = dirnode;
-	    objdirmtime = xmlXPathEvalExpression (BAD_CAST "@mtime", xpath);
-	    
-	    node = objdirmtime->nodesetval->nodeTab[0];
-	    dirmtime = xmlNodeListGetString (manindex_xml, 
-	                                     node->xmlChildrenNode, 1);
-	    
-	    objdirname = xmlXPathEvalExpression (BAD_CAST "name[1]", xpath);
+            lang     = xmlGetProp (dirnode, BAD_CAST "lang");
+	    dirmtime = xmlGetProp (dirnode, BAD_CAST "mtime");
 
-	    node = objdirname->nodesetval->nodeTab[0];
-	    dirname = xmlNodeListGetString (manindex_xml, 
-	                                    node->xmlChildrenNode, 1);
+	    if (lang == NULL)
+		lang = xmlStrdup (BAD_CAST "C");
+
+	    for (node = dirnode->children; node != NULL; node = node->next) {
+		if (node->type == XML_ELEMENT_NODE && 
+		    g_str_equal ((gchar *)node->name, "name")) {
+		    dirname = xmlNodeGetContent (node);
+		}
+	    }
+
+	    g_hash_table_insert (priv->man_dirlang, g_strdup ((gchar *)dirname), 
+	                         g_strdup ((gchar *)lang));
 	    
 	    /* if we can't stat the dirname for some reason, then skip adding
 	     * this directory to the TOC, remove the dirnode, and set the flag
@@ -1148,25 +1172,22 @@ create_toc_from_index (YelpTocPager *pager, gchar *index_file)
 	    /* otherwise just read from the index file */
 	    } else {
 	    
-		objmanpages = xmlXPathEvalExpression (BAD_CAST "page", xpath);
-
-		for (k=0; k < objmanpages->nodesetval->nodeNr; k++) {
-		    xmlNodePtr node = objmanpages->nodesetval->nodeTab[k];
+		for (node = dirnode->children; node != NULL; node = node->next) {
 		    xmlChar *manpage = NULL;
 
-		    manpage = xmlNodeListGetString (manindex_xml,
-		                                    node->xmlChildrenNode, 1);
-
-		    add_man_page_to_toc (pager, (gchar *)dirname, (gchar *)manpage);
-		    priv->manpage_count++;
-		    xmlFree (manpage);
-	        }
-		xmlXPathFreeObject (objmanpages);
+		    if (node->type == XML_ELEMENT_NODE && 
+		        g_str_equal ((gchar *)node->name, "page")) {
+			manpage = xmlNodeGetContent (node);
+			add_man_page_to_toc (pager, (gchar *)dirname, (gchar *)manpage);
+			priv->manpage_count++;
+			xmlFree (manpage);
+		    }
+		}
+	    
 	    }
+	    xmlFree (lang);
 	    xmlFree (dirmtime);
 	    xmlFree (dirname);
-	    xmlXPathFreeObject (objdirmtime);
-	    xmlXPathFreeObject (objdirname);
 	}
 
 	/* cleanup */
@@ -1209,7 +1230,7 @@ process_mandir_pending (YelpTocPager *pager)
 					 XML_PARSE_NOBLANKS | XML_PARSE_NOCDATA  |
 					 XML_PARSE_NOENT    | XML_PARSE_NOERROR  |
 					 XML_PARSE_NONET    );
-	priv->man_secthash = g_hash_table_new (g_str_hash, g_str_equal);
+	priv->man_secthash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
 
 	xpath = xmlXPathNewContext (priv->toc_doc);
 	obj = xmlXPathEvalExpression (BAD_CAST "//toc", xpath);
@@ -1222,11 +1243,14 @@ process_mandir_pending (YelpTocPager *pager)
 		gchar **sects = g_strsplit ((gchar *)sect, " ", 0);
 
 		for (j = 0; sects[j] != NULL; j++)
-		    g_hash_table_insert (priv->man_secthash, sects[j], node);
+		    g_hash_table_insert (priv->man_secthash, 
+					 g_strdup (sects[j]), node);
+		g_strfreev (sects);
 	    }
-
+	    xmlFree (sect);
 	    xml_trim_titles (node);
 	}
+	xmlXPathFreeObject (obj);
 	xmlXPathFreeContext (xpath);
     }
 
@@ -1244,6 +1268,10 @@ process_mandir_pending (YelpTocPager *pager)
 	gchar **manpaths = NULL;
 	gchar **mandirs = NULL;
 
+	if (!priv->man_dirlang)
+	    priv->man_dirlang = g_hash_table_new_full (g_str_hash, g_str_equal,
+	                                               g_free,     g_free);
+
 	/* check for the existence of the xml cache file in ~/.gnome2/yelp.d/
 	 * if it exists, use it as the source for man pages instead of 
 	 * searching the hard disk for them - should make startup much faster */
@@ -1258,6 +1286,7 @@ process_mandir_pending (YelpTocPager *pager)
 	mandirs = yelp_get_man_paths ();
 	priv->manindex_xml = xmlNewDoc (BAD_CAST "1.0");
 	priv->root = xmlNewNode (NULL, BAD_CAST "manindex");
+	xmlNewProp (priv->root, BAD_CAST "lang", BAD_CAST langs[0]);
 	priv->ins  = priv->root;
 
 	xmlDocSetRootElement (priv->manindex_xml, priv->root);
@@ -1275,6 +1304,33 @@ process_mandir_pending (YelpTocPager *pager)
 	manpaths = g_strsplit (manpath, G_SEARCHPATH_SEPARATOR_S, -1);
 	g_free (manpath);
 
+	/* order is important here, since we may encounter collisions when adding 
+	 * a man page to the hash table "man_manhash".
+	 *
+	 * We want the resulting list to be sorted in order of priority:
+	 * 1) Highest priority is given to the base directory name returned 
+	 *    from $MANPATH or the manpath program.  So if that is 
+	 *    /usr/local/share:/usr/share then the man page 
+	 *    /usr/local/share/man/man1/python.1.gz would have precedence over
+	 *    /usr/share/man/man1/python.1.gz 
+	 * 2) Next highest priority is given to the language.
+	 * 3) Lowest priority is given to the section name, i.e. "man0p" "man1"
+	 *    since a man page from one section should never conflict with a man page
+	 *    from another section (because of the .0p or .1 section in the filename)
+	 *
+	 * Here is an example of how it should be sorted when 
+	 *   $MANPATH=/usr/local/share:/usr/share and LANGUAGE="es"
+	 *   
+	 * Adding /usr/local/share/man/es/man0p
+	 * Adding /usr/share/man/es/man0p
+	 * Adding /usr/local/share/man/man0p
+	 * Adding /usr/share/man/man0p
+	 * Adding /usr/local/share/man/es/man1
+	 * Adding /usr/share/man/es/man1
+	 * Adding /usr/local/share/man/man1
+	 * Adding /usr/share/man/man1 
+	 */
+	
 	for (i=0; mandirs[i] != NULL; i++) {
 	    GSList *tmplist = NULL;
 
@@ -1287,12 +1343,35 @@ process_mandir_pending (YelpTocPager *pager)
 			                            mandirs[i],
 			                            NULL);
 
+		    g_hash_table_insert (priv->man_dirlang, g_strdup (dirname),
+		                         g_strdup (langs[j]));
+
+		    /* prepend to list for speed, reverse it at the end */
 		    tmplist = g_slist_prepend (tmplist, dirname);
 		}
 	    }
+	    tmplist = g_slist_reverse (tmplist);
 
 	    priv->mandir_list = g_slist_prepend (priv->mandir_list, tmplist);
 	}
+	priv->mandir_list = g_slist_reverse (priv->mandir_list);
+
+#ifdef YELP_DEBUG
+	/* debugging: print out lists in order */
+	GSList *list1 = priv->mandir_list;
+	GSList *list2 = NULL;
+
+	while (list1 != NULL) {
+	    list2 = list1->data;
+	    while (list2 && list2->data) {
+	        debug_print (DB_INFO, "Dir=%s\n", (gchar *)list2->data);
+	        list2 = g_slist_next (list2);
+	    }
+	    list1 = g_slist_next (list1);
+	}
+#endif
+
+	g_strfreev (manpaths);
  
 	priv->mandir_ptr = priv->mandir_list;
 	if (priv->mandir_list && priv->mandir_list->data) {
@@ -1311,9 +1390,10 @@ process_mandir_pending (YelpTocPager *pager)
 	if ((dir = g_dir_open (dirname, 0, NULL))) {
 	    struct stat buf;
 	    gchar mtime_str[20];
+	    gchar *lang = NULL;
 
 	    if (g_stat (dirname, &buf) < 0)
-		g_warning ("Unable to stat dir: \"%s\"\n", dirname);
+		debug_print (DB_WARN, "Unable to stat dir: \"%s\"\n", dirname);
 
 	    /* check to see if we need create a new hash table */
 	    if (!priv->man_manhash) {
@@ -1327,7 +1407,17 @@ process_mandir_pending (YelpTocPager *pager)
 
 	    g_snprintf (mtime_str, 20, "%u", (guint) buf.st_mtime);
 
+	    lang = g_hash_table_lookup (priv->man_dirlang, dirname);
+
 	    priv->ins = xmlNewChild (priv->ins, NULL, BAD_CAST "dir", NULL);
+
+	    if (lang == NULL) {
+		g_print ("dir %s is null\n", dirname);
+		lang = "C";
+	    }
+	    
+	    xmlNewProp (priv->ins, BAD_CAST "lang", BAD_CAST lang);
+
 	    xmlNewProp (priv->ins, BAD_CAST "mtime", BAD_CAST mtime_str);
 	    xmlAddChild (priv->ins, xmlNewText (BAD_CAST "\n      "));
 	    xmlNewChild (priv->ins, NULL, BAD_CAST "name", BAD_CAST dirname);
@@ -1424,7 +1514,8 @@ process_info_pending (YelpTocPager *pager)
 	    xmlNodePtr tmp;
 	    xmlNodePtr new_node = NULL;
 	    gboolean menufound = FALSE;
-	    
+	    gchar ** amp;
+
 	    if (!priv->toc_doc) {
 		xmlXPathContextPtr xpath;
 		xmlXPathObjectPtr  obj;
@@ -1436,12 +1527,22 @@ process_info_pending (YelpTocPager *pager)
 		xpath = xmlXPathNewContext (priv->toc_doc);
 		obj = xmlXPathEvalExpression (BAD_CAST "//toc", xpath);
 		node = obj->nodesetval->nodeTab[0];
+		xmlXPathFreeObject (obj);
+		xmlXPathFreeContext (xpath);
 	    }
 	    
 	    channel = yelp_io_channel_new_file (filename, NULL);
 	    g_io_channel_read_to_end (channel, &str, &len, NULL);
 	    g_io_channel_shutdown (channel, FALSE, NULL);
-	    
+	    g_io_channel_unref (channel);
+	    /* Check the string for the special character '&' which
+	     * screws up the xml.  If its there, it needs escaping
+	     */
+	    amp = g_strsplit (str, "&", -1);
+	    g_free (str);
+	    str = g_strjoinv ("&amp;", amp);
+	    g_strfreev (amp);
+
 	    files = g_strsplit (str, "\n", -1);
 	    
 	    for (ptr = files; *ptr != NULL; ptr++) {
@@ -1473,6 +1574,7 @@ process_info_pending (YelpTocPager *pager)
 			p = g_strdup (part2);
 			p = g_strstrip (p);
 			tooltip = g_strdup_printf (_("Read info page for %s"), p);
+			g_free (p);
 			path = part1+1;
 			part1 = strchr (path, ')');
 			if (!part1)
@@ -1525,15 +1627,20 @@ process_info_pending (YelpTocPager *pager)
 		    } else if (!g_ascii_isspace (**ptr) && g_ascii_isprint(**ptr)){
 			new_node = g_hash_table_lookup (categories, *ptr);
 			if (!new_node) {
+			    gchar *tmp;
 
 			    /* A new section */
 			    static int sectno = 1;
 			    new_node = xmlNewChild (node, NULL, BAD_CAST "toc",
 						    NULL);
+			    tmp = g_strdup_printf ("%d", sectno);
 			    xmlNewNsProp (new_node, NULL, BAD_CAST "sect",
-					  BAD_CAST g_strdup_printf ("%d", sectno));
+					  BAD_CAST tmp);
+			    g_free (tmp);
+			    tmp = g_strdup_printf ("infosect%d", sectno);
 			    xmlNewNsProp (new_node, NULL, BAD_CAST "id",
-					  BAD_CAST g_strdup_printf ("infosect%d", sectno));
+					  BAD_CAST tmp);
+			    g_free (tmp);
 			    sectno++;
 			    xmlNewChild (new_node, NULL, BAD_CAST "title",
 					 BAD_CAST *ptr);
@@ -1782,12 +1889,32 @@ process_cleanup (YelpTocPager *pager)
 {
     YelpTocPagerPriv *priv = pager->priv;
 
+#ifdef ENABLE_MAN
+    /* clean up the man directory language hash table */
+    if (priv->man_dirlang) {
+        g_hash_table_destroy (priv->man_dirlang);
+        priv->man_dirlang = NULL;
+    }
+
+    /* clean up the man page section hash table */
+    if (priv->man_secthash) {
+	g_hash_table_destroy (priv->man_secthash);
+	priv->man_secthash = NULL;
+    }
+#endif
+
     /* cleanup the stylesheet used to process the toc */
     if (priv->stylesheet) {
 	xsltFreeStylesheet (priv->stylesheet);
 	priv->stylesheet = NULL;
     }
 
+    /* cleanup the parser context */
+    if (priv->parser) {
+	xmlFreeParserCtxt (priv->parser);
+	priv->parser = NULL;
+    }
+
     /* we only ever want to run this function once, so always return false */
     return FALSE;
 }
