@@ -48,6 +48,7 @@
 #include "yelp-toc-pager.h"
 #include "yelp-window.h"
 #include "yelp-print.h"
+#include "yelp-debug.h"
 
 #ifdef ENABLE_MAN
 #include "yelp-man-pager.h"
@@ -60,12 +61,6 @@
 #include "gtkentryaction.h"
 #endif
 
-#ifdef YELP_DEBUG
-#define d(x) x
-#else
-#define d(x)
-#endif
-
 #define YELP_CONFIG_WIDTH          "/yelp/Geometry/width"
 #define YELP_CONFIG_HEIGHT         "/yelp/Geometry/height"
 #define YELP_CONFIG_WIDTH_DEFAULT  "600"
@@ -82,6 +77,9 @@ typedef struct {
     YelpDocInfo *doc_info;
 
     gchar *frag_id;
+    GtkWidget *menu_entry;
+    YelpWindow *window;
+    gint callback;
 
     gchar *page_title;
     gchar *frag_title;
@@ -196,6 +194,10 @@ static void               history_step_back      (YelpWindow       *window);
 static YelpHistoryEntry * history_pop_back       (YelpWindow       *window);
 static YelpHistoryEntry * history_pop_forward    (YelpWindow       *window);
 static void               history_entry_free     (YelpHistoryEntry *entry);
+static void               history_back_to        (GtkMenuItem      *menuitem,
+						  YelpHistoryEntry *entry);
+static void               history_forward_to     (GtkMenuItem      *menuitem,
+						  YelpHistoryEntry *entry);
 
 static void               load_data_free         (YelpLoadData     *data);
 
@@ -213,7 +215,8 @@ static gboolean    window_find_entry_key_pressed_cb (GtkWidget       *widget,
 						     gpointer         data);
 static void        window_find_save_settings      (YelpWindow        *window);
 static void        window_find_buttons_set_sensitive (YelpWindow      *window,
-						      gboolean        sensitive);
+						      gboolean        prev,
+						      gboolean        next);
 static void        window_find_clicked_cb         (GtkWidget         *button,
 						   YelpWindow        *window);
 static void        window_find_next_cb            (GtkAction *action, 
@@ -265,6 +268,8 @@ struct _YelpWindowPriv {
     gchar          *current_frag;
     GSList         *history_back;
     GSList         *history_forward;
+    GtkWidget      *back_menu;
+    GtkWidget      *forward_menu;
 
     /* Callbacks and Idles */
     gulong          start_handler;
@@ -570,6 +575,7 @@ history_push_back (YelpWindow *window)
     YelpWindowPriv   *priv;
     YelpHistoryEntry *entry;
     GtkAction        *action;
+    gchar            *title;
 
     g_return_if_fail (YELP_IS_WINDOW (window));
     g_return_if_fail (window->priv->current_doc != NULL);
@@ -586,6 +592,26 @@ history_push_back (YelpWindow *window)
     action = gtk_action_group_get_action (priv->action_group, "GoBack");
     if (action)
 	g_object_set (G_OBJECT (action), "sensitive", TRUE, NULL);
+    title = gtk_window_get_title (window);
+
+    if (g_str_equal (title, "Loading..."))
+	entry->page_title = g_strdup ("Unknown Page");
+    else
+	entry->page_title = g_strdup (title);
+
+    entry->window = window;
+    
+    entry->menu_entry = gtk_menu_item_new_with_label (entry->page_title);
+    g_object_ref (entry->menu_entry);
+
+    entry->callback = g_signal_connect (G_OBJECT (entry->menu_entry), 
+		      "activate", 
+		      G_CALLBACK (history_back_to),
+		      entry);
+		      
+
+    gtk_menu_shell_prepend (window->priv->back_menu, entry->menu_entry);
+    gtk_widget_show (entry->menu_entry);
 }
 
 static void
@@ -594,6 +620,7 @@ history_push_forward (YelpWindow *window)
     YelpWindowPriv   *priv;
     YelpHistoryEntry *entry;
     GtkAction        *action;
+    gchar            *title;
 
     g_return_if_fail (YELP_IS_WINDOW (window));
     g_return_if_fail (window->priv->current_doc != NULL);
@@ -610,6 +637,26 @@ history_push_forward (YelpWindow *window)
     action = gtk_action_group_get_action (priv->action_group, "GoForward");
     if (action)
 	g_object_set (G_OBJECT (action), "sensitive", TRUE, NULL);
+
+    title = gtk_window_get_title (window);
+
+    if (g_str_equal (title, "Loading..."))
+	entry->page_title = g_strdup ("Unknown Page");
+    else
+	entry->page_title = g_strdup (title);
+
+    entry->window = window;
+    entry->menu_entry = gtk_menu_item_new_with_label (entry->page_title);
+    g_object_ref (entry->menu_entry);
+
+    entry->callback = g_signal_connect (G_OBJECT (entry->menu_entry), 
+		      "activate", 
+		      G_CALLBACK (history_forward_to),
+		      entry);
+		      
+
+    gtk_menu_shell_prepend (window->priv->forward_menu, entry->menu_entry);
+    gtk_widget_show (entry->menu_entry);
 }
 
 static void
@@ -725,9 +772,108 @@ history_entry_free (YelpHistoryEntry *entry)
     g_free (entry->page_title);
     g_free (entry->frag_title);
 
+    if (entry->menu_entry) {
+	gtk_widget_destroy (entry->menu_entry);
+    }
+
     g_free (entry);
 }
 
+static void 
+history_back_to (GtkMenuItem *menuitem, YelpHistoryEntry *entry)
+{
+    YelpWindow       *window;
+    YelpWindowPriv   *priv;
+    YelpHistoryEntry *latest;
+
+    window = entry->window;
+
+    priv = window->priv;
+
+    latest = history_pop_back (window);
+    history_push_forward (window);
+
+    while (!g_str_equal (latest->page_title, entry->page_title)) {
+	priv->history_forward = g_slist_prepend (priv->history_forward, latest);
+	gtk_container_remove (priv->back_menu, latest->menu_entry);
+	
+	g_signal_handler_disconnect (latest->menu_entry, latest->callback);
+	latest->callback = g_signal_connect (G_OBJECT (latest->menu_entry), 
+					     "activate", 
+					     G_CALLBACK (history_forward_to),
+					     latest);
+	
+	
+	gtk_menu_shell_prepend (window->priv->forward_menu, latest->menu_entry);
+	gtk_widget_show (latest->menu_entry);
+
+	
+	latest = history_pop_back (window);
+    }
+    if (latest->menu_entry) {
+	gtk_widget_destroy (latest->menu_entry);
+	latest->menu_entry = NULL;
+    }
+
+    if (priv->current_doc)
+	yelp_doc_info_unref (priv->current_doc);
+    if (priv->current_frag)
+	g_free (priv->current_frag);
+
+    priv->current_doc  = yelp_doc_info_ref (entry->doc_info);
+    priv->current_frag = g_strdup (entry->frag_id);
+
+    window_do_load (window, entry->doc_info, entry->frag_id);
+
+}
+
+static void
+history_forward_to (GtkMenuItem *menuitem, YelpHistoryEntry *entry)
+{
+    YelpWindow       *window;
+    YelpWindowPriv   *priv;
+    YelpHistoryEntry *latest;
+
+    window = entry->window;
+
+    priv = window->priv;
+
+    latest = history_pop_forward (window);
+    history_push_back (window);
+
+    while (!g_str_equal (latest->page_title, entry->page_title)) {
+	priv->history_back = g_slist_prepend (priv->history_back, latest);
+	gtk_container_remove (priv->forward_menu, latest->menu_entry);
+	
+	g_signal_handler_disconnect (latest->menu_entry, latest->callback);
+	latest->callback = g_signal_connect (G_OBJECT (latest->menu_entry), 
+					     "activate", 
+					     G_CALLBACK (history_back_to),
+					     latest);
+
+	gtk_menu_shell_prepend (window->priv->back_menu, latest->menu_entry);
+	gtk_widget_show (latest->menu_entry);
+
+	
+	latest = history_pop_forward (window);
+    }
+    if (latest->menu_entry) {
+	gtk_widget_destroy (latest->menu_entry);
+	latest->menu_entry = NULL;
+    }
+
+    if (priv->current_doc)
+	yelp_doc_info_unref (priv->current_doc);
+    if (priv->current_frag)
+	g_free (priv->current_frag);
+
+    priv->current_doc  = yelp_doc_info_ref (entry->doc_info);
+    priv->current_frag = g_strdup (entry->frag_id);
+
+    window_do_load (window, entry->doc_info, entry->frag_id);
+
+}
+
 static void
 load_data_free (YelpLoadData *data)
 {
@@ -762,8 +908,8 @@ yelp_window_load (YelpWindow *window, const gchar *uri)
     } else {
 	real_uri = g_strdup (uri);
     }
-    d (g_print ("yelp_window_load\n"));
-    d (g_print ("  uri = \"%s\"\n", uri));
+    debug_print (DB_FUNCTION, "entering\n");
+    debug_print (DB_ARG, "  uri = \"%s\"\n", uri);
 
     if (!real_uri) {
 	GError *error = NULL;
@@ -851,7 +997,7 @@ window_do_load (YelpWindow  *window,
     g_return_if_fail (YELP_IS_WINDOW (window));
     g_return_if_fail (doc_info != NULL);
 
-    d (g_print ("window_do_laod\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     action = gtk_action_group_get_action (window->priv->action_group, 
 					  "PrintDocument");
@@ -934,7 +1080,7 @@ window_do_load (YelpWindow  *window,
 	window_error (window, error, TRUE);
     }
 
-    window_find_buttons_set_sensitive (window, TRUE);    
+    window_find_buttons_set_sensitive (window, TRUE, TRUE);    
 }
 
 /******************************************************************************/
@@ -992,7 +1138,21 @@ search_activated (GtkAction *action,
 
     search_terms = gtk_entry_action_get_text (GTK_ENTRY_ACTION (action));
 
-    uri = encode_search_uri (search_terms);
+    /* Do some basic checking here - if someones looking for a man
+     * or info page, save doing the search and just bring them directly to
+     * the relevant page
+     * Trigger it using "man:<foo>", "man <foo>" or "info:<foo>"
+     */
+    if (g_str_has_prefix (search_terms, "man:") || 
+	g_str_has_prefix (search_terms, "info:")) {
+	uri = g_strdup (search_terms);
+    } else if (g_str_has_prefix (search_terms, "man ") && 
+	       !strstr (&(search_terms[4])," ")) {
+	uri = g_strdup (search_terms);
+	uri[3]=':';
+    } else {
+	uri = encode_search_uri (search_terms);
+    }
 
     yelp_window_load (window, uri);
 
@@ -1009,6 +1169,11 @@ window_populate (YelpWindow *window)
     GError            *error = NULL;
     GtkAction         *action;
     GtkTreeSelection  *selection;
+    GtkWidget         *toolbar;
+    GtkWidget         *toolitem;
+    GtkWidget         *b_proxy;
+    GtkWidget         *f_proxy;
+
 
     priv = window->priv;
 
@@ -1022,6 +1187,27 @@ window_populate (YelpWindow *window)
     gtk_action_group_add_actions (priv->action_group,
 				  entries, G_N_ELEMENTS (entries),
 				  window);
+
+    b_proxy = GTK_WIDGET (gtk_menu_tool_button_new (NULL, NULL));
+
+    priv->back_menu = gtk_menu_new ();
+
+    priv->forward_menu = gtk_menu_new ();
+
+    f_proxy = GTK_WIDGET (gtk_menu_tool_button_new (NULL, NULL));
+
+    gtk_menu_tool_button_set_menu (GTK_MENU_TOOL_BUTTON (b_proxy), 
+				   priv->back_menu);
+
+    action = gtk_action_group_get_action(priv->action_group, "GoBack");
+
+    gtk_action_connect_proxy (action, b_proxy);
+
+    action = gtk_action_group_get_action (priv->action_group, "GoForward");
+    gtk_action_connect_proxy (action, f_proxy);
+
+    gtk_menu_tool_button_set_menu (GTK_MENU_TOOL_BUTTON (f_proxy), 
+				   priv->forward_menu);
     
 #ifdef ENABLE_SEARCH
     action =  gtk_entry_action_new ("Search",
@@ -1059,8 +1245,6 @@ window_populate (YelpWindow *window)
 
     yelp_bookmarks_register (window);
 
-    gtk_ui_manager_ensure_update (priv->ui_manager);
-
     action = gtk_action_group_get_action (priv->action_group, "GoBack");
     if (action)
 	g_object_set (G_OBJECT (action), "sensitive", FALSE, NULL);
@@ -1068,6 +1252,18 @@ window_populate (YelpWindow *window)
     if (action)
 	g_object_set (G_OBJECT (action), "sensitive", FALSE, NULL);
 
+    gtk_ui_manager_ensure_update (priv->ui_manager);
+
+    toolbar = gtk_ui_manager_get_widget(priv->ui_manager, "ui/tools");
+
+    gtk_toolbar_insert (GTK_TOOLBAR (toolbar), f_proxy, 0);
+    gtk_toolbar_insert (GTK_TOOLBAR (toolbar), b_proxy, 0);
+
+    /*To keep the seperator but get rid of the extra widget*/
+    toolitem = gtk_ui_manager_get_widget(priv->ui_manager, "ui/tools/GoFor");
+
+    gtk_container_remove (toolbar, toolitem);
+
     priv->popup = gtk_ui_manager_get_widget(priv->ui_manager, "ui/main_popup");
     priv->maillink = gtk_ui_manager_get_widget(priv->ui_manager, "ui/mail_popup");
     priv->merge_id = gtk_ui_manager_new_merge_id (priv->ui_manager);
@@ -1574,10 +1770,10 @@ window_handle_page (YelpWindow   *window,
     priv = window->priv;
     window_disconnect (window);
 
-    d (g_print ("window_handle_page\n"));
-    d (g_print ("  page->page_id  = \"%s\"\n", page->page_id));
-    d (g_print ("  page->title    = \"%s\"\n", page->title));
-    d (g_print ("  page->contents = %i bytes\n", strlen (page->contents)));
+    debug_print (DB_FUNCTION, "entering\n");
+    debug_print (DB_ARG, "  page->page_id  = \"%s\"\n", page->page_id);
+    debug_print (DB_ARG, "  page->title    = \"%s\"\n", page->title);
+    debug_print (DB_ARG, "  page->contents = %i bytes\n", strlen (page->contents));
 
     model = gtk_tree_view_get_model (GTK_TREE_VIEW (priv->side_sects));
     pager = yelp_doc_info_get_pager (priv->current_doc);
@@ -1598,12 +1794,17 @@ window_handle_page (YelpWindow   *window,
 	    if (yelp_pager_page_contains_frag (pager,
 					       id,
 					       priv->current_frag)) {
-		GtkTreePath *path = gtk_tree_model_get_path (model, &iter);
-
-		gtk_tree_view_expand_to_path (GTK_TREE_VIEW (priv->side_sects),
-					      path);
+		GtkTreePath *path = NULL;
+		GtkTreeIter parent;
+		if (gtk_tree_model_iter_parent (model, &parent, &iter)) {
+		    path = gtk_tree_model_get_path (model, &parent);
+		    gtk_tree_view_expand_to_path (GTK_TREE_VIEW (priv->side_sects),
+						  path);
+		    gtk_tree_path_free(path);
+		}
+		path = gtk_tree_model_get_path (model, &iter);
 		gtk_tree_selection_select_path (selection, path);
-
+		
 		gtk_tree_path_free (path);
 		g_free (id);
 		break;
@@ -1650,7 +1851,7 @@ window_handle_page (YelpWindow   *window,
 				 priv->current_frag,
 				 YELP_URI_TYPE_FILE);
 
-    d (g_print ("  uri            = %s\n", uri));
+    debug_print (DB_ARG, "  uri            = %s\n", uri);
 
     yelp_html_set_base_uri (priv->html_view, uri);
     yelp_html_open_stream (priv->html_view, "application/xhtml+xml");
@@ -1752,8 +1953,8 @@ pager_start_cb (YelpPager   *pager,
     page_id = yelp_pager_resolve_frag (pager,
 				       window->priv->current_frag);
 
-    d (g_print ("pager_start_cb\n"));
-    d (g_print ("  page_id=\"%s\"\n", page_id));
+    debug_print (DB_FUNCTION, "entering\n");
+    debug_print (DB_ARG, "  page_id=\"%s\"\n", page_id);
 
     window_set_sections (window,
 			 yelp_pager_get_sections (pager));
@@ -1781,8 +1982,8 @@ pager_page_cb (YelpPager *pager,
     YelpWindow  *window = YELP_WINDOW (user_data);
     YelpPage    *page;
 
-    d (g_print ("pager_page_cb\n"));
-    d (g_print ("  page_id=\"%s\"\n", page_id));
+    debug_print (DB_FUNCTION, "entering\n");
+    debug_print (DB_ARG, "  page_id=\"%s\"\n", page_id);
 
     if (yelp_pager_page_contains_frag (pager,
 				       page_id,
@@ -1810,7 +2011,7 @@ pager_error_cb (YelpPager   *pager,
     YelpWindow *window = YELP_WINDOW (user_data);
     GError *error = yelp_pager_get_error (pager);
 
-    d (g_print ("pager_error_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     window_disconnect (window);
     window_error (window, error, TRUE);
@@ -1823,7 +2024,7 @@ pager_cancel_cb (YelpPager   *pager,
 		gpointer     user_data)
 {
     YelpWindow *window = YELP_WINDOW (user_data);
-    d (g_print ("pager_cancel_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     window_disconnect (window);
 }
@@ -1835,7 +2036,7 @@ pager_finish_cb (YelpPager   *pager,
     GError *error = NULL;
     YelpWindow  *window = YELP_WINDOW (user_data);
 
-    d (g_print ("pager_finish_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     window_disconnect (window);
 
@@ -1860,8 +2061,8 @@ html_uri_selected_cb (YelpHtml  *html,
 {
     YelpWindow *window = YELP_WINDOW (user_data);
 
-    d (g_print ("html_uri_selected_cb\n"));
-    d (g_print ("  uri = \"%s\"\n", uri));
+    debug_print (DB_FUNCTION, "entering\n");
+    debug_print (DB_ARG, "  uri = \"%s\"\n", uri);
 
     if (!handled) {
 	yelp_window_load (window, uri);
@@ -1968,7 +2169,7 @@ tree_drag_data_get_cb (GtkWidget         *widget,
 
     g_return_if_fail (YELP_IS_WINDOW (window));
 
-    d (g_print ("tree_drag_data_get_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     priv = window->priv;
 
@@ -2030,7 +2231,7 @@ typedef struct {
 static void
 print_disconnect (PrintStruct *data)
 {
-    d(g_print ("print disconnect\n"));
+    debug_print (DB_FUNCTION, "entering\n");
     if (data->page_handler) {
 	g_signal_handler_disconnect (data->pager,
 				     data->page_handler);
@@ -2061,8 +2262,8 @@ print_pager_page_cb (YelpPager *pager,
 {
     PrintStruct *data = user_data;
     YelpPage    *page;
-    d (g_print ("print_pager_page_cb\n"));
-    d (g_print ("  page_id=\"%s\"\n", page_id));
+    debug_print (DB_FUNCTION, "entering\n");
+    debug_print (DB_ARG, "  page_id=\"%s\"\n", page_id);
 
     page = (YelpPage *) yelp_pager_get_page (pager, page_id);
 
@@ -2072,7 +2273,7 @@ print_pager_page_cb (YelpPager *pager,
 	int length, offset;
 	char *uri;
 	GtkWidget *vbox = gtk_vbox_new (FALSE, FALSE);
-	d(g_print (page->contents));
+	debug_print (DB_DEBUG, page->contents);
 
 	gtk_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
 	html = yelp_html_new ();
@@ -2088,14 +2289,14 @@ print_pager_page_cb (YelpPager *pager,
 				     page_id,
 				     YELP_URI_TYPE_FILE);
 
-	d (g_print ("  uri            = %s\n", uri));
+	debug_print (DB_ARG, "  uri            = %s\n", uri);
 
 	yelp_html_set_base_uri (html, uri);
 	g_free (uri);
 
 	yelp_html_open_stream (html, "application/xhtml+xml");
 	for (length = strlen (page->contents), offset = 0; length > 0; length -= BUFFER_SIZE, offset += BUFFER_SIZE) {
-	    d(g_print ("data: %.*s\n", MIN (length, BUFFER_SIZE), page->contents + offset));
+	    debug_print (DB_DEBUG, "data: %.*s\n", MIN (length, BUFFER_SIZE), page->contents + offset);
 	    yelp_html_write (html, page->contents + offset, MIN (length, BUFFER_SIZE));
 	}
 	yelp_html_close (html);
@@ -2114,7 +2315,7 @@ print_pager_error_cb (YelpPager   *pager,
     PrintStruct *data = user_data;
     /*     GError *error = yelp_pager_get_error (pager);*/
 
-    d (g_print ("print_pager_error_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     print_disconnect (data);
 }
@@ -2124,7 +2325,7 @@ print_pager_cancel_cb (YelpPager   *pager,
 		       gpointer     user_data)
 {
     PrintStruct *data = user_data;
-    d (g_print ("print_pager_cancel_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     print_disconnect (data);
 }
@@ -2135,7 +2336,7 @@ print_pager_finish_cb (YelpPager   *pager,
 {
     PrintStruct *data = user_data;
 
-    d (g_print ("print_pager_finish_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     print_disconnect (data);
 }
@@ -2215,7 +2416,7 @@ window_print_page_cb (GtkAction *action, YelpWindow *window)
 	
 	yelp_html_open_stream (html, "application/xhtml+xml");
 	for (length = strlen (page->contents), offset = 0; length > 0; length -= BUFFER_SIZE, offset += BUFFER_SIZE) {
-	    d (g_print ("data: %.*s\n", MIN (length, BUFFER_SIZE), page->contents + offset));
+	    debug_print (DB_DEBUG, "data: %.*s\n", MIN (length, BUFFER_SIZE), page->contents + offset);
 	    yelp_html_write (html, page->contents + offset, MIN (length, BUFFER_SIZE));
 	}
 	yelp_html_close (html);
@@ -2386,7 +2587,7 @@ window_find_cb (GtkAction *action, YelpWindow *window)
 
     priv = window->priv;
 
-    d (g_print ("window_find_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     gtk_widget_show_all (priv->find_bar);
     gtk_widget_grab_focus (priv->find_entry);
@@ -2408,7 +2609,7 @@ window_reload_cb (GtkAction *action, YelpWindow *window)
 
     g_return_if_fail (YELP_IS_WINDOW (window));
 
-    d (g_print ("window_reload_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     if (window->priv->current_doc) {
 	YelpLoadData *data;
@@ -2514,7 +2715,7 @@ window_go_forward_cb (GtkAction *action, YelpWindow *window)
 static void
 window_go_home_cb (GtkAction *action, YelpWindow *window)
 {
-    d (g_print ("window_go_home_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
     g_return_if_fail (YELP_IS_WINDOW (window));
     yelp_window_load (window, "x-yelp-toc:");
 }
@@ -2525,7 +2726,7 @@ window_go_previous_cb (GtkAction *action, YelpWindow *window)
     YelpWindowPriv *priv;
     gchar *base, *uri;
 
-    d (g_print ("window_go_previous_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
     g_return_if_fail (YELP_IS_WINDOW (window));
     g_return_if_fail (window->priv->current_doc);
     priv = window->priv;
@@ -2545,7 +2746,7 @@ window_go_next_cb (GtkAction *action, YelpWindow *window)
     YelpWindowPriv *priv;
     gchar *base, *uri;
 
-    d (g_print ("window_go_next_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
     g_return_if_fail (YELP_IS_WINDOW (window));
     g_return_if_fail (window->priv->current_doc);
     priv = window->priv;
@@ -2565,7 +2766,7 @@ window_go_toc_cb (GtkAction *action, YelpWindow *window)
     YelpWindowPriv *priv;
     gchar *base, *uri;
 
-    d (g_print ("window_go_toc_cb\n"));
+    debug_print (DB_FUNCTION, "entering\n");
     g_return_if_fail (YELP_IS_WINDOW (window));
     g_return_if_fail (window->priv->current_doc);
     priv = window->priv;
@@ -2585,7 +2786,7 @@ window_add_bookmark_cb (GtkAction *action, YelpWindow *window)
     gchar *uri;
     YelpWindowPriv *priv = window->priv;
 
-    d (g_print ("window_add_bookmark\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     uri = yelp_doc_info_get_uri (priv->current_doc, priv->current_frag,
 				 YELP_URI_TYPE_NO_FILE);
@@ -2714,8 +2915,8 @@ location_response_cb (GtkDialog *dialog, gint id, YelpWindow *window)
 {
     g_return_if_fail (YELP_IS_WINDOW (window));
 
-    d (g_print ("location_response_cb\n"));
-    d (g_print ("  id = %i\n", id));
+    debug_print (DB_FUNCTION, "entering\n");
+    debug_print (DB_ARG, "  id = %i\n", id);
 
     if (id == GTK_RESPONSE_OK) {
 	const gchar *uri = 
@@ -2763,12 +2964,6 @@ window_find_again (YelpWindow *window, gboolean forward)
 {
     YelpWindowPriv *priv = window->priv;
 
-#ifndef TYPEAHEADFIND
-    window_find_save_settings (window);
-
-    yelp_html_set_find_props (priv->html_view, priv->find_string, FALSE, FALSE);
-#endif /* TYPEAHEADFIND */
-
     return yelp_html_find_again (priv->html_view, forward);
 }
 
@@ -2779,14 +2974,9 @@ window_find_previous_cb (GtkAction *action, YelpWindow *window)
 
     priv = window->priv;
     if (!window_find_again (window, FALSE)) {
-	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), TRUE);
-	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), FALSE);
-	g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", TRUE, 
-		      NULL);
-	g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", FALSE, 
-		      NULL);
+	window_find_buttons_set_sensitive (window, FALSE, TRUE);
     } else {
-	window_find_buttons_set_sensitive (window, TRUE);
+	window_find_buttons_set_sensitive (window, TRUE, TRUE);
     }
 }
 
@@ -2797,14 +2987,9 @@ window_find_next_cb (GtkAction *action, YelpWindow *window)
 
     priv = window->priv;
     if (!window_find_again (window, TRUE)) {
-	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), FALSE);
-	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), TRUE);
-	g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", FALSE, 
-		      NULL);
-	g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", TRUE, 
-		      NULL);
+	window_find_buttons_set_sensitive (window, TRUE, FALSE);
     } else {
-	window_find_buttons_set_sensitive (window, TRUE);
+	window_find_buttons_set_sensitive (window, TRUE, TRUE);
     }
 }
 
@@ -2820,26 +3005,16 @@ window_find_clicked_cb (GtkWidget  *widget,
 
     if (GTK_TOOL_ITEM (widget) == priv->find_next) {
 	if (!window_find_again (window, TRUE)) {
-	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), FALSE);
-	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), TRUE);
-	    g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", FALSE, 
-			  NULL);
-	    g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", TRUE, 
-			  NULL);
+	    window_find_buttons_set_sensitive (window, TRUE, FALSE);
 	} else {
-	    window_find_buttons_set_sensitive (window, TRUE);
+	    window_find_buttons_set_sensitive (window, TRUE, TRUE);
 	}
     }
     else if (GTK_TOOL_ITEM (widget) == priv->find_prev) {
 	if (!window_find_again (window, FALSE)) {
-	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), TRUE);
-	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), FALSE);
-	    g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", TRUE, 
-			  NULL);
-	    g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", FALSE, 
-			  NULL);
+	    window_find_buttons_set_sensitive (window, FALSE, TRUE);
 	} else {
-	    window_find_buttons_set_sensitive (window, TRUE);
+	    window_find_buttons_set_sensitive (window, TRUE, TRUE);
 	}
     }
     else
@@ -2865,17 +3040,12 @@ window_find_entry_changed_cb (GtkEditable *editable,
     found = window_find_action (window);
 
     if (text == NULL || text[0] == '\0')
-	window_find_buttons_set_sensitive (window, FALSE);
+	window_find_buttons_set_sensitive (window, FALSE, FALSE);
     else
 	if (found) {
-	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), TRUE);
-	    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), FALSE);
-	    g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", TRUE, 
-			  NULL);
-	    g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", FALSE, 
-			  NULL);
+	    window_find_buttons_set_sensitive (window, FALSE, TRUE);
 	} else {
-	    window_find_buttons_set_sensitive (window, TRUE);
+	    window_find_buttons_set_sensitive (window, TRUE, TRUE);
 	}
  
     g_free (text);
@@ -2893,18 +3063,11 @@ window_find_entry_activate_cb (GtkEditable *editable,
     window = YELP_WINDOW (data);
     priv = window->priv;
 
-    window_find_buttons_set_sensitive (window, FALSE);
-
     /* search forward and update the buttons */
     if (!window_find_again (window, TRUE)) {
-	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), FALSE);
-	gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), TRUE);
-	g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", FALSE, 
-		      NULL);
-	g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", TRUE, 
-		      NULL);
+	window_find_buttons_set_sensitive (window, TRUE, FALSE);
     } else {
-	window_find_buttons_set_sensitive (window, TRUE);
+	window_find_buttons_set_sensitive (window, TRUE, TRUE);
     }
 }
 
@@ -2916,14 +3079,14 @@ window_find_entry_key_pressed_cb (GtkWidget    *widget,
     YelpWindow     *window;
     YelpWindowPriv *priv;
 
+    debug_print (DB_FUNCTION, "entering\n");
+
     if (key->keyval == GDK_Escape) {
 	g_return_val_if_fail (YELP_IS_WINDOW (data), FALSE);
 
 	window = YELP_WINDOW (data);
 	priv = window->priv;
 
-	d (g_print ("window_find_entry_key_pressed_cb\n"));
-
 	gtk_widget_hide_all (priv->find_bar);
 
 	return TRUE;
@@ -2934,17 +3097,17 @@ window_find_entry_key_pressed_cb (GtkWidget    *widget,
 
 static void
 window_find_buttons_set_sensitive (YelpWindow  *window,
-				   gboolean    sensitive)
+				   gboolean    prev, gboolean next)
 {
     YelpWindowPriv *priv;
 
     priv = window->priv;
 
-    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), sensitive);
-    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), sensitive);
-    g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", sensitive, 
+    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_next), next);
+    gtk_widget_set_sensitive (GTK_WIDGET (priv->find_prev), prev);
+    g_object_set (G_OBJECT (priv->find_next_menu), "sensitive", next, 
 		  NULL);
-    g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", sensitive, 
+    g_object_set (G_OBJECT (priv->find_prev_menu), "sensitive", prev, 
 		  NULL);
 }
 
@@ -2999,15 +3162,15 @@ idle_write (IdleWriterContext *context)
     g_return_val_if_fail (context != NULL, FALSE);
     g_return_val_if_fail (context->window != NULL, FALSE);
 
-    d (g_print ("idle_write\n"));
+    debug_print (DB_FUNCTION, "entering\n");
 
     priv = context->window->priv;
 
     switch (context->type) {
     case IDLE_WRITER_MEMORY:
-	d (g_print ("  context->buffer = %i bytes\n", strlen (context->buffer)));
-	d (g_print ("  context->cur    = %i\n", context->cur));
-	d (g_print ("  context->length = %i\n", context->length));
+	debug_print (DB_DEBUG, "  context->buffer = %i bytes\n", strlen (context->buffer));
+	debug_print (DB_DEBUG, "  context->cur    = %i\n", context->cur);
+	debug_print (DB_DEBUG, "  context->length = %i\n", context->length);
 
 	if (context->cur + BUFFER_SIZE < context->length) {
 	    yelp_html_write (priv->html_view,
